/*
 *  Rexx/CDT - Mix C/Database Toolchest
 *  Copyright (C) 1999-2003 Mark Hessling <M.Hessling@qut.edu.au>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

static char RCSid[] = "$Id: rxcdt.c,v 1.3 2003/01/11 06:05:28 mark Exp $";

#define REXXCDT_VERSION "1.1"
#define REXXCDT_DATE    "28 Dec 2002"

#include "rxcdt.h"
#include "rxpack.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "isam.h"
#include "isamerr.h"
#include "cbtree.h"

static char *RxPackageName = "rxcdt";


RexxFunctionHandler CDTaddrec;
RexxFunctionHandler CDTclose_db;
RexxFunctionHandler CDTcreate_db;
RexxFunctionHandler CDTdelrec;
RexxFunctionHandler CDTfindhead;
RexxFunctionHandler CDTfindkey;
RexxFunctionHandler CDTfindnext;
RexxFunctionHandler CDTfindprev;
RexxFunctionHandler CDTfindtail;
RexxFunctionHandler CDTgetindexdesc;
RexxFunctionHandler CDTgeterror;
RexxFunctionHandler CDTgetfldcount;
RexxFunctionHandler CDTgetfldnames;
RexxFunctionHandler CDTgetfldtype;
RexxFunctionHandler CDTgetindexnames;
RexxFunctionHandler CDTgetrec;
RexxFunctionHandler CDTopen_index;
RexxFunctionHandler CDTcreate_index;
RexxFunctionHandler CDTadd_index;
RexxFunctionHandler CDTmodrec;
RexxFunctionHandler CDTopen_db;
RexxFunctionHandler CDTisamexit;
RexxFunctionHandler CDTisaminit;
RexxFunctionHandler CDTLoadFuncs;
RexxFunctionHandler CDTDropFuncs;
RexxFunctionHandler CDTVariable;

/*-----------------------------------------------------------------------------
 * Table of Mix CDT Functions. Used to install/de-install functions.
 * If you change this table, don't forget to change the table at the end
 * of this file.
 *----------------------------------------------------------------------------*/
RexxFunction RexxCDTFunctions[] = {
   { "CDTADDREC"          , CDTaddrec           ,"CDTaddrec"         , 1 },
   { "CDTCLOSE_DB"        , CDTclose_db         ,"CDTclose_db"       , 1 },
   { "CDTCREATE_DB"       , CDTcreate_db        ,"CDTcreate_db"      , 1 },
   { "CDTDELREC"          , CDTdelrec           ,"CDTdelrec"         , 1 },
   { "CDTFINDHEAD"        , CDTfindhead         ,"CDTfindhead"       , 1 },
   { "CDTFINDKEY"         , CDTfindkey          ,"CDTfindkey"        , 1 },
   { "CDTFINDNEXT"        , CDTfindnext         ,"CDTfindnext"       , 1 },
   { "CDTFINDPREV"        , CDTfindprev         ,"CDTfindprev"       , 1 },
   { "CDTFINDTAIL"        , CDTfindtail         ,"CDTfindtail"       , 1 },
   { "CDTGETINDEXDESC"    , CDTgetindexdesc     ,"CDTgetindexdesc"   , 1 },
   { "CDTGETERROR"        , CDTgeterror         ,"CDTgeterror"       , 1 },
   { "CDTGETFLDCOUNT"     , CDTgetfldcount      ,"CDTgetfldcount"    , 1 },
   { "CDTGETFLDNAMES"     , CDTgetfldnames      ,"CDTgetfldnames"    , 1 },
   { "CDTGETFLDTYPE"      , CDTgetfldtype       ,"CDTgetfldtype"     , 1 },
   { "CDTGETINDEXNAMES"   , CDTgetindexnames    ,"CDTgetindexnames"  , 1 },
   { "CDTGETREC"          , CDTgetrec           ,"CDTgetrec"         , 1 },
   { "CDTOPEN_INDEX"      , CDTopen_index       ,"CDTopen_index"     , 1 },
   { "CDTCREATE_INDEX"    , CDTcreate_index     ,"CDTcreate_index"   , 1 },
   { "CDTADD_INDEX"       , CDTadd_index        ,"CDTadd_index"      , 1 },
   { "CDTMODREC"          , CDTmodrec           ,"CDTmodrec"         , 1 },
   { "CDTOPEN_DB"         , CDTopen_db          ,"CDTopen_db"        , 1 },
   { "CDTISAMEXIT"        , CDTisamexit         ,"CDTisamexit"       , 1 },
   { "CDTISAMINIT"        , CDTisaminit         ,"CDTisaminit"       , 1 },
   { "CDTLOADFUNCS"       , CDTLoadFuncs        ,"CDTLoadFuncs"      , 0 },/* Don't load for DLL */
   { "CDTDROPFUNCS"       , CDTDropFuncs        ,"CDTDropFuncs"      , 1 },
   { "CDTVARIABLE"        , CDTVariable         ,"CDTVariable"       , 1 },
   { NULL, NULL, NULL,0 }
};


#define ARGLEN(x)      (argv[x].strlength)
#define ARGPTR(x)      (argv[x].strptr)
#define ARGTERM(x)     (argv[x].strptr[argv[x].strlength] = '\0')

#define PRXSTRLEN(x)   ((x)->strptr ? (x)->strlength : 0UL)
#define PRXSTRPTR(x)   ((x)->strptr)
#define PRXSTRTERM(x)  ((x)->strptr[(x)->strlength] = '\0')

#define RXSTRLEN(x)    ((x).strptr ? (x).strlength : 0UL)
#define RXSTRPTR(x)    ((x).strptr)
#define RXSTRTERM(x)   ((x).strptr[(x).strlength] = '\0')

static REXXCDTDATA RexxCDTData = { 2, "", NULL };

/*
 * These functions used by loader.c to obtain package-specific info
 */
int RexxCDTInitialiser( RxPackageGlobalDataDef *MyGlob )
{
   RexxCDTData.RxPackageGlobalData = MyGlob;
   return 0;
}

int RexxCDTTerminator( RxPackageGlobalDataDef *MyGlob )
{
   return 0;
}

RexxSubcomHandler *getRexxCDTSubcomHandler( void )
{
   return NULL;
}

RexxFunction *getRexxCDTFunctions( void )
{
   return RexxCDTFunctions;
}

PackageInitialiser *getRexxCDTInitialiser( void )
{
   return RexxCDTInitialiser;
}

PackageTerminator *getRexxCDTTerminator( void )
{
   return NULL;
}

void free_data( int num_fields, char **recfields )
{
   int i;
   for ( i = 0; i < num_fields; i++ )
   {
      free( recfields[i] );
   }
   free( recfields );
}

/*-----------------------------------------------------------------------------
 * This function 
 *----------------------------------------------------------------------------*/
void RxSetDecimals

#ifdef HAVE_PROTO
   ( int decimals )
#else
   ( decimals )
   int decimals;
#endif

{
   InternalTrace( RexxCDTData.RxPackageGlobalData, "RxSetDecimals", "%d", decimals );
   RexxCDTData.RxDecimals = decimals;
}

/*-----------------------------------------------------------------------------
 * This function 
 *----------------------------------------------------------------------------*/
int RxGetDecimals

#ifdef HAVE_PROTO
   ( void )
#else
   ( )
#endif

{  
   InternalTrace( RexxCDTData.RxPackageGlobalData, "RxGetDecimals", NULL );
   return ( RexxCDTData.RxDecimals );
}

/*-----------------------------------------------------------------------------
 * This function 
 *----------------------------------------------------------------------------*/
void RxSetInternalError

#ifdef HAVE_PROTO
    (char *fmt, ...)
#else
    (fmt)
    char *fmt;
#endif

{
   va_list argptr;

   va_start( argptr, fmt );
   vsprintf( RexxCDTData.RxInternalError, fmt, argptr );
   va_end( argptr );
   return;
}

/*-----------------------------------------------------------------------------
 * This function 
 *----------------------------------------------------------------------------*/
char *RxGetInternalError

#ifdef HAVE_PROTO
    (void)
#else
    ()
#endif

{
   return( RexxCDTData.RxInternalError );
}
/*-----------------------------------------------------------------------------
 * This function 
 *----------------------------------------------------------------------------*/
void RxClearInternalError

#ifdef HAVE_PROTO
    (void)
#else
    ()
#endif

{
   strcpy( RexxCDTData.RxInternalError, "" );
   return;
}

char **Stem2CArray( char *stem, int *num_fields, int *fields, Db_Obj *db )
{
   int rc=I_OK,i,fldnum;
   char **recfields;
   RXSTRING rxs, *prxs;
   short i_integer, *p_integer;
   unsigned short i_unsigned, *p_unsigned;
   long i_long, *p_long;
   float i_float, *p_float;
   double i_double, *p_double;
   char *tmp;

   if ( ( recfields = (char **)malloc( *num_fields * sizeof(char *) ) ) == NULL )
   {
      *num_fields = 0;
      RxSetInternalError( "Memory allocation error" );
      return NULL;
   }
   for ( i = 0; i < *num_fields; i++ )
   {
      prxs = GetRexxVariable( RexxCDTData.RxPackageGlobalData,  stem, &rxs, i+1 );
      if ( prxs == NULL )
      {
         RxSetInternalError( "Error getting Rexx variable: %s%d", stem, i+1 );
         free_data( i, recfields );
         *num_fields = 0;
         return NULL;
      }
      RXSTRTERM( rxs );
      fldnum = fields[i];
      switch( igetfldtype( db, fldnum ) )
      {
         case I_STRING:
         case I_CSTRING:
            recfields[i] = RXSTRPTR(rxs);
            break;
         case I_INTEGER:
            i_integer = (short)strtol( RXSTRPTR(rxs), &tmp, 10 );
            if ( *RXSTRPTR(rxs) != 0 && *tmp == 0)
            {
               p_integer = (short *)malloc( sizeof(short) );
               *p_integer = i_integer;
               recfields[i] = (char *)p_integer;
               free( RXSTRPTR(rxs) );
               break;
            }
            RxSetInternalError( "Invalid INTEGER: %s", RXSTRPTR(rxs) );
            rc = I_ERROR;
            break;
         case I_UNSIGNED:
            i_unsigned = (unsigned short)strtol( RXSTRPTR(rxs), &tmp, 10 );
            if ( *RXSTRPTR(rxs) != 0 && *tmp == 0)
            {
               p_unsigned = (unsigned short *)malloc( sizeof(unsigned short) );
               *p_unsigned = i_unsigned;
               recfields[i] = (char *)p_unsigned;
               free( RXSTRPTR(rxs) );
               break;
            }
            RxSetInternalError( "Invalid UNSIGNED: %s", RXSTRPTR(rxs) );
            rc = I_ERROR;
            break;
         case I_LONG:
            i_long = strtol( RXSTRPTR(rxs), &tmp, 10 );
            if ( *RXSTRPTR(rxs) != 0 && *tmp == 0)
            {
               p_long = (long *)malloc( sizeof(long) );
               *p_long = i_long;
               recfields[i] = (char *)p_long;
               free( RXSTRPTR(rxs) );
               break;
            }
            RxSetInternalError( "Invalid LONG: %s", RXSTRPTR(rxs) );
            rc = I_ERROR;
            break;
         case I_FLOAT:
            i_float = (float)strtod( RXSTRPTR(rxs), &tmp );
            if ( *RXSTRPTR(rxs) != 0 && *tmp == 0)
            {
               p_float = (float *)malloc( sizeof(float) );
               *p_float = i_float;
               recfields[i] = (char *)p_float;
               free( RXSTRPTR(rxs) );
               break;
            }
            RxSetInternalError( "Invalid FLOAT: %s", RXSTRPTR(rxs) );
            rc = I_ERROR;
            break;
         case I_DOUBLE:
            i_double = strtod( RXSTRPTR(rxs), &tmp );
            if ( *RXSTRPTR(rxs) != 0 && *tmp == 0)
            {
               p_double = (double *)malloc( sizeof(double) );
               *p_double = i_double;
               recfields[i] = (char *)p_double;
               free( RXSTRPTR(rxs) );
               break;
            }
            RxSetInternalError( "Invalid DOUBLE: %s", RXSTRPTR(rxs) );
            rc = I_ERROR;
            break;
         case I_BINARY:
            if ( igetfldlen(db, i) != RXSTRLEN(rxs) )
            {
               rc = I_ERROR;
               break;
            }
            recfields[i] = RXSTRPTR(rxs);
            break;
         default:
            rc = I_ERROR;
            break;
      }
      if ( rc == I_ERROR )
      {
         free_data( i, recfields );
         *num_fields = 0;
         free( RXSTRPTR(rxs) );
         return NULL;
      }
   }
   return recfields;
}

static Db_Obj *arg2db( RXSTRING *argv )
{
   argv->strptr[argv->strlength] = '\0';
   return( (Db_Obj*)atol( argv->strptr ) );
}

static Index_Obj *arg2idx( RXSTRING *argv )
{
   argv->strptr[argv->strlength] = '\0';
   return( (Index_Obj*)atol( argv->strptr ) );
}

static int arg2int( RXSTRING *argv )
{
   argv->strptr[argv->strlength] = '\0';
   return( atoi( argv->strptr ) );
}

static int RxReturnState( RXSTRING *retstr, int rc )
{
   InternalTrace( RexxCDTData.RxPackageGlobalData, "RxReturnState", "%x,%d", retstr, rc );

   switch( rc )
   {
      case I_OK:
         strcpy( (char *)retstr->strptr, "OK" );
         break;
      case I_ERROR:
         strcpy( (char *)retstr->strptr, "ERROR" );
         break;
      case I_NOTFOUND:
         strcpy( (char *)retstr->strptr, "NOTFOUND" );
         break;
      case I_FOUND:
         strcpy( (char *)retstr->strptr, "FOUND" );
         break;
      case I_EOI:
         strcpy( (char *)retstr->strptr, "EOI" );
         break;
      case I_BOI:
         strcpy( (char *)retstr->strptr, "BOI" );
         break;
      default:
         strcpy( (char *)retstr->strptr, "?" );
         break;
   }
   retstr->strlength = strlen( (char *)retstr->strptr );
   if ( RexxCDTData.RxPackageGlobalData
   &&   RexxCDTData.RxPackageGlobalData->RxRunFlags & MODE_VERBOSE )
   {
      (void)fprintf( RexxCDTData.RxPackageGlobalData->RxTraceFilePointer,
         "++ Exit %s with value \"%s\"\n",
         RexxCDTData.RxPackageGlobalData->FName,
         retstr->strptr );
      fflush( RexxCDTData.RxPackageGlobalData->RxTraceFilePointer );
   }
   return( 0 );
}

static int RxReturnType( Db_Obj *db, int fldnum, RXSTRING *retstr, int rc )
{
   int binfldlen;

   InternalTrace( RexxCDTData.RxPackageGlobalData, "RxReturnType", "%x,%d,%x,%ld", db, fldnum, retstr, rc );
   switch( rc )
   {
      case I_STRING:
         strcpy( (char *)retstr->strptr, "S" );
         break;
      case I_CSTRING:
         strcpy( (char *)retstr->strptr, "C" );
         break;
      case I_INTEGER:
         strcpy( (char *)retstr->strptr, "I" );
         break;
      case I_UNSIGNED:
         strcpy( (char *)retstr->strptr, "U" );
         break;
      case I_LONG:
         strcpy( (char *)retstr->strptr, "L" );
         break;
      case I_FLOAT:
         strcpy( (char *)retstr->strptr, "F" );
         break;
      case I_DOUBLE:
         strcpy( (char *)retstr->strptr, "D" );
         break;
      case I_BINARY:
         binfldlen = igetfldlen( db, fldnum );
         sprintf( (char *)retstr->strptr, "B%d",binfldlen );
         break;
      default:
         strcpy( (char *)retstr->strptr, "?" );
         break;    
   }
   retstr->strlength = strlen( (char *)retstr->strptr );
   if ( RexxCDTData.RxPackageGlobalData
   &&   RexxCDTData.RxPackageGlobalData->RxRunFlags & MODE_VERBOSE )
   {
      (void)fprintf( RexxCDTData.RxPackageGlobalData->RxTraceFilePointer,
         "++ Exit %s with value \"%s\"\n",
         RexxCDTData.RxPackageGlobalData->FName,
         retstr->strptr );
      fflush( RexxCDTData.RxPackageGlobalData->RxTraceFilePointer );
   }
   return( 0 );
}

/*====== Here come the real interface functions to cdt ======*/

RFH_RETURN_TYPE CDTaddrec
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{ 
   int i,rc,num_fields;
   Db_Obj *db;
   Index_Obj *idx;
   char **recfields;
   RXSTRING rxs, *prxs;
   int *fields;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 3, 3 ) )
      return( 1 );
   db = arg2db( argv );
   if ( ARGLEN(1) == 0 )
      idx = NULL;
   else
      idx = arg2idx( argv+1 ) ;
   ARGTERM(2);
   /*
    * Allocate space for field pointers.
    */
   num_fields = igetfldcount( db );
   fields = (int *)malloc( num_fields*sizeof(int) );
   if ( fields == NULL )
   {
      RxSetInternalError( "Memory allocation error" );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   for ( i = 0; i < num_fields; i++ )
      fields[i] = i;
   prxs = GetRexxVariable( RexxCDTData.RxPackageGlobalData,  ARGPTR(2), &rxs, 0 );
   if ( prxs == NULL )
   {
      free( fields );
      RxSetInternalError( "Error getting Rexx variable: %s%d", ARGPTR(2), 0 );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   RXSTRTERM( rxs );
   if ( atoi( RXSTRPTR(rxs) ) != num_fields )
   {
      free( fields );
      free( RXSTRPTR(rxs) );
      RxSetInternalError( "Number of fields in stem: %s, does not match number of fields in database file: %d", RXSTRPTR(rxs), num_fields );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   free( RXSTRPTR(rxs) );
   rc = I_OK;
   recfields = Stem2CArray( ARGPTR(2), &num_fields, fields, db );
   free( fields );
   if ( recfields == NULL )
      rc = I_ERROR;
   else
   {
      rc = iaddrec( db, idx, recfields );
   }
   free_data( num_fields, recfields );
   return RxReturnState( retstr, rc ) ;
}

RFH_RETURN_TYPE CDTclose_db
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int rc;
   Db_Obj *db;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 1, 1 ) )
      return( 1 );
   db = arg2db( argv );
   rc = iclose_db( db );
   return RxReturnState( retstr, rc ) ;
}

RFH_RETURN_TYPE CDTcreate_db
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   Db_Obj *db;
   int blksize,num_fields,i;
   char **fielddesc;
   RXSTRING rxs, *prxs;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 3, 3 ) )
      return( 1 );
   ARGTERM(0);
   blksize = arg2int( argv+1 );
   prxs = GetRexxVariable( RexxCDTData.RxPackageGlobalData,  ARGPTR(2), &rxs, 0 );
   if ( prxs == NULL )
   {
      RxSetInternalError( "Error getting Rexx variable: %s%d", ARGPTR(2), 0 );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   RXSTRTERM( rxs );
   if ( ( num_fields = atoi( RXSTRPTR(rxs) ) ) == 0 )
   {
      free( RXSTRPTR(rxs) );
      RxSetInternalError( "Number of fields in stem: %s, must be a whole number > 0", RXSTRPTR(rxs) );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   free( RXSTRPTR(rxs) );
   /*
    * Allocate space for field pointers.
    */
   if ( ( fielddesc = (char **)malloc( ( num_fields + 1 ) * sizeof(char *) ) ) == NULL )
   {
      RxSetInternalError( "Memory allocation error" );
      return RxReturnState( retstr, I_ERROR ) ;
   }

   for ( i = 0; i < num_fields; i++ )
   {
      prxs = GetRexxVariable( RexxCDTData.RxPackageGlobalData,  ARGPTR(2), &rxs, i+1 );
      if ( prxs == NULL )
      {
         free( fielddesc );
         free( RXSTRPTR(rxs) );
         RxSetInternalError( "Error getting Rexx variable: %s%d", ARGPTR(2), i+1 );
         return RxReturnState( retstr, I_ERROR ) ;
      }
      RXSTRTERM( rxs );
      fielddesc[i] = RXSTRPTR(rxs);
   }
   fielddesc[i] = NULL;
   db = icreate_db( ARGPTR(0), blksize, fielddesc );
   for ( i = 0; i < num_fields; i++ )
      free( fielddesc[i] );
   free( fielddesc );
   return RxReturnNumber( RexxCDTData.RxPackageGlobalData, retstr, (long)db ) ;
}
   
RFH_RETURN_TYPE CDTdelrec
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int rc;
   Db_Obj *db;
   Index_Obj *idx;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 2, 2 ) )
      return( 1 );
   db = arg2db( argv );
   idx = arg2idx( argv+1 );
   rc = idelrec( db, idx );
   return RxReturnState( retstr, rc ) ;
}


RFH_RETURN_TYPE CDTfindhead
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int rc;
   Db_Obj *db;
   Index_Obj *idx;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 2, 2 ) )
      return( 1 );
   db = arg2db( argv );
   idx = arg2idx( argv+1 );
   rc = ifindhead( db, idx );
   return RxReturnState( retstr, rc ) ;
}

RFH_RETURN_TYPE CDTfindkey
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int i,rc,num_fields;
   Db_Obj *db;
   Index_Obj *idx;
   char **recfields;
   RXSTRING rxs, *prxs;
   int *fields;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 3, 3 ) )
      return( 1 );
   db = arg2db( argv );
   idx = arg2idx( argv+1 ) ;
   ARGTERM(2);
   num_fields = igetdesc_count( db, idx );
   prxs = GetRexxVariable( RexxCDTData.RxPackageGlobalData,  ARGPTR(2), &rxs, 0 );
   if ( prxs == NULL )
   {
      RxSetInternalError( "Error getting Rexx variable: %s%d", ARGPTR(2), 0 );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   RXSTRTERM( rxs );
   if ( atoi( RXSTRPTR(rxs) ) != num_fields )
   {
      free( RXSTRPTR(rxs) );
      RxSetInternalError( "Number of fields in stem: %s, does not match number of fields in index: %d", RXSTRPTR(rxs), num_fields );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   free( RXSTRPTR(rxs) );
   /*
    * Allocate space for field pointers.
    */
   fields = (int *)malloc( num_fields*sizeof(int) );
   if ( fields == NULL )
   {
      RxSetInternalError( "Memory allocation error" );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   for ( i = 0; i < num_fields; i++ )
   {
      fields[i] = igetdesc_fldnbr( db, idx, i );
   }
   rc = I_OK;
   recfields = Stem2CArray( ARGPTR(2), &num_fields, fields, db );
   free( fields );
   if ( recfields == NULL )
      rc = I_ERROR;
   else
   {
      rc = ifindkey( db, idx, recfields );
   }
   free_data( num_fields, recfields );
   return RxReturnState( retstr, rc ) ;
}

RFH_RETURN_TYPE CDTfindnext
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int rc;
   Db_Obj *db;
   Index_Obj *idx;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 2, 2 ) )
      return( 1 );
   db = arg2db( argv );
   idx = arg2idx( argv+1 );
   rc = ifindnext( db, idx );
   return RxReturnState( retstr, rc ) ;
}

RFH_RETURN_TYPE CDTfindprev
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int rc;
   Db_Obj *db;
   Index_Obj *idx;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 2, 2 ) )
      return( 1 );
   db = arg2db( argv );
   idx = arg2idx( argv+1 );
   rc = ifindprev( db, idx );
   return RxReturnState( retstr, rc ) ;
}

RFH_RETURN_TYPE CDTfindtail
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int rc;
   Db_Obj *db;
   Index_Obj *idx;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 2, 2 ) )
      return( 1 );
   db = arg2db( argv );
   idx = arg2idx( argv+1 );
   rc = ifindtail( db, idx );
   return RxReturnState( retstr, rc ) ;
}

RFH_RETURN_TYPE CDTgetindexdesc
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int i,rc;
   Db_Obj *db;
   Index_Obj *idx;
   char **desc;
   char *ptr;
   char ptr1[10];

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 3, 3 ) )
      return( 1 );
   db = arg2db( argv );
   idx = arg2idx( argv+1 );
   ARGTERM(2);
   desc = igetdesc( db, idx );
   if (desc)
   {
      if ( ( ptr = (char *)malloc( ARGLEN(2) + 10 ) ) == NULL )
      {
         RxSetInternalError( "Memory allocation error" );
         rc = I_ERROR;
      }
      else
      {
         rc = I_OK;
         for ( i = 0; desc[i] != NULL; i++ )
         {
            sprintf( ptr, "%s%d", ARGPTR(2), i+1 );
            make_upper( ptr );
            if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr, strlen( ptr ), desc[i], strlen( desc[i] ) ) )
            {
               RxSetInternalError( "Error setting Rexx variable: %s", ptr );
               rc = I_ERROR;
               break;
            }
         }
         if ( rc != I_ERROR )
         {
            sprintf( ptr, "%s0", ARGPTR(2) );
            make_upper( ptr );
            sprintf( ptr1, "%d", i );
            if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr, strlen( ptr ), ptr1, strlen( ptr1 ) ) )
            {
               RxSetInternalError( "Error setting Rexx variable: %s", ptr );
               rc = I_ERROR;
            }
         }
         free( ptr );
      }
   }
   else
   {
      rc = I_ERROR;
   }
   return RxReturnState( retstr, rc ) ;
}


RFH_RETURN_TYPE CDTgeterror
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int rc;
   char *ptr;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 1, 1 ) )
      return( 1 );
   ARGTERM(0);
   if ( ( ptr = (char *)malloc( ARGLEN(0) + 20 ) ) == NULL )
   {
      rc = I_ERROR;
   }
   else
   {
      sprintf( ptr, "%sINTERNAL", ARGPTR(0) );
      make_upper( ptr );
      if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr, strlen( ptr ), RxGetInternalError(), strlen( RxGetInternalError() ) ) )
      {
         return RxReturnState( retstr, I_ERROR ) ;
      }
      sprintf( ptr, "%sCDTMSG", ARGPTR(0) );
      make_upper( ptr );
      if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr, strlen( ptr ), isammsg( isam_errno ), strlen( isammsg( isam_errno ) ) ) )
      {
         return RxReturnState( retstr, I_ERROR ) ;
      }
      if ( isam_errno == I_CBT )
      {
         sprintf( ptr, "%sCBERRMSG", ARGPTR(0) );
         make_upper( ptr );
         if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr, strlen( ptr ), cberrmsg( cberrcode ), strlen( cberrmsg( cberrcode ) ) ) )
         {
            return RxReturnState( retstr, I_ERROR ) ;
         }
         sprintf( ptr, "%sCBFCNMSG", ARGPTR(0) );
         make_upper( ptr );
         if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr, strlen( ptr ), cbfcnmsg( cbfcncode ), strlen( cbfcnmsg( cbfcncode ) ) ) )
         {
            return RxReturnState( retstr, I_ERROR ) ;
         }
         sprintf( ptr, "%sCBCONMSG", ARGPTR(0) );
         make_upper( ptr );
         if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr, strlen( ptr ), cbconmsg( cbconcode ), strlen( cbconmsg( cbconcode ) ) ) )
         {
            return RxReturnState( retstr, I_ERROR ) ;
         }
      }
      else
      {
         sprintf( ptr, "%sCBERRMSG", ARGPTR(0) );
         make_upper( ptr );
         if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr, strlen( ptr ), "", 0 ) )
         {
            return RxReturnState( retstr, I_ERROR ) ;
         }
         sprintf( ptr, "%sCBFCNMSG", ARGPTR(0) );
         make_upper( ptr );
         if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr, strlen( ptr ), "", 0 ) )
         {
            return RxReturnState( retstr, I_ERROR ) ;
         }
         sprintf( ptr, "%sCBCONMSG", ARGPTR(0) );
         make_upper( ptr );
         if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr, strlen( ptr ), "", 0 ) )
         {
            return RxReturnState( retstr, I_ERROR ) ;
         }
      }
      free( ptr );
   }
   return RxReturnState( retstr, I_OK ) ;
}

RFH_RETURN_TYPE CDTgetfldcount
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int rc;
   Db_Obj *db;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 1, 1 ) )
      return( 1 );
   db = arg2db( argv );
   rc = igetfldcount( db );
   if ( rc == I_ERROR )
      return RxReturnState( retstr, rc ) ;
   else
      return RxReturnNumber( RexxCDTData.RxPackageGlobalData, retstr, rc ) ;
}

RFH_RETURN_TYPE CDTgetfldnames
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int i,rc;
   Db_Obj *db;
   char **desc;
   char *ptr;
   char buf1[20];

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 2, 2 ) )
      return( 1 );
   db = arg2db( argv );
   ARGTERM(1);
   desc = igetfldnames( db );
   if (desc)
   {
      if ( ( ptr = (char *)malloc( ARGLEN(1) + 10 ) ) == NULL )
      {
         RxSetInternalError( "Memory allocation error" );
         rc = I_ERROR;
      }
      else
      {
         rc = I_OK;
         for ( i = 0; desc[i] != NULL; i++ )
         {
            sprintf( ptr, "%s%d", ARGPTR(1), i+1 );
            make_upper( ptr );
            if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr, strlen( ptr ), desc[i], strlen( desc[i] ) ) )
            {
               RxSetInternalError( "Error setting Rexx variable: %s", ptr );
               rc = I_ERROR;
               break;
            }
         }
         if ( rc != I_ERROR )
         {
            sprintf( ptr, "%s0", ARGPTR(1) );
            make_upper( ptr );
            sprintf( buf1, "%d", i );
            if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr, strlen( ptr ), buf1, strlen( buf1 ) ) )
            {
               RxSetInternalError( "Error setting Rexx variable: %s", ptr );
               rc = I_ERROR;
            }
         }
         free( ptr );
      }
   }
   else
   {
      rc = I_ERROR;
   }
   return RxReturnState( retstr, rc ) ;
}

RFH_RETURN_TYPE CDTgetfldtype
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int rc, num, num_fields;
   Db_Obj *db;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 2, 2 ) )
      return( 1 );
   db = arg2db( argv );
   num = ( arg2int( argv+1 ) ) - 1;
   num_fields = igetfldcount( db );
   if ( num < 0 || num >= num_fields )
   {
      RxSetInternalError( "Number of fields in stem: %s, must be less than number of fields in database file: %d", ARGPTR(1), num_fields );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   rc = igetfldtype( db, num );
   if ( rc == I_ERROR )
      return RxReturnState( retstr, rc ) ;
   else
      return RxReturnType( db, num, retstr, rc ) ;
}

RFH_RETURN_TYPE CDTgetindexnames
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int i,rc;
   Db_Obj *db;
   char **desc;
   char *ptr;
   char buf1[20];

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 2, 2 ) )
      return( 1 );
   db = arg2db( argv );
   ARGTERM(1);
   desc = igetindexnames( db );
   if (desc)
   {
      if ( ( ptr = (char *)malloc( ARGLEN(1) + 10 ) ) == NULL )
      {
         RxSetInternalError( "Memory allocation error" );
         rc = I_ERROR;
      }
      else
      {
         rc = I_OK;
         for ( i = 0; desc[i] != NULL; i++ )
         {
            sprintf( ptr, "%s%d", ARGPTR(1), i+1 );
            make_upper( ptr );
            if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr, strlen( ptr ), desc[i], strlen( desc[i] ) ) )
            {
               RxSetInternalError( "Error setting Rexx variable: %s", ptr );
               rc = I_ERROR;
               break;
            }
         }
         if ( rc != I_ERROR )
         {
            sprintf( ptr, "%s0", ARGPTR(1) );
            make_upper( ptr );
            sprintf( buf1, "%d", i );
            if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr, strlen( ptr ), buf1, strlen( buf1 ) ) )
            {
               RxSetInternalError( "Error setting Rexx variable: %s", ptr );
               rc = I_ERROR;
            }
         }
         free( ptr );
      }
   }
   else
   {
      rc = I_ERROR;
   }
   return RxReturnState( retstr, rc ) ;
}

RFH_RETURN_TYPE CDTgetrec
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int i,rc,num_fields,value_len;
   Db_Obj *db;
   Index_Obj *idx;
   char **fields;
   char *ptr,*ptr1,*value;
   short i_integer;
   unsigned short i_unsigned;
   long i_long;
   float i_float;
   double i_double;
   Rec_Len reclength;
   char buf[50];
   char buf1[20];

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 3, 3 ) )
      return( 1 );
   db = arg2db( argv );
   idx = arg2idx( argv+1 );
   ARGTERM(2);
   /*
    * Allocate space for record buffer.
    */
   if ( igetreclen( db, idx, &reclength ) == I_ERROR )
      return RxReturnState( retstr, I_ERROR ) ;
   if ( ( ptr = (char *)malloc( reclength ) ) == NULL )
   {
      RxSetInternalError( "Memory allocation error" );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   if ( ( ptr1 = (char *)malloc( ARGLEN(2) + 10 ) ) == NULL )
   {
      free( ptr );
      RxSetInternalError( "Memory allocation error" );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   /*
    * Allocate space for field pointers.
    */
   num_fields = igetfldcount( db );
   if ( ( fields = (char **)malloc( ( num_fields + 1 ) * sizeof(char *) ) ) == NULL )
   {
      free( ptr );
      RxSetInternalError( "Memory allocation error" );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   /*
    * Get the record
    */
   rc = igetrec( db, idx, fields, ptr, reclength );
   if ( rc == I_ERROR )
   {
      free( ptr );
      free( ptr1 );
      free( fields );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   /*
    * If the record is EOI or BOI, set the number of fields to
    * zero
    */
   if ( rc == I_EOI
   ||   rc == I_BOI )
   {
      num_fields = 0;
   }
   /*
    * For each field in the record extract it and set a rexx variable
    */
   for ( i = 0; i < num_fields; i++ )
   {
      switch( igetfldtype( db, i ) )
      {
         case I_STRING:
         case I_CSTRING:
            value = fields[i];
            value_len = strlen( fields[i] );
            break;
         case I_INTEGER:
            memcpy( &i_integer, fields[i], sizeof(i_integer) );
            sprintf( buf, "%d", i_integer );
            value = buf;
            value_len = strlen( buf );
            break;
         case I_UNSIGNED:
            memcpy( &i_unsigned, fields[i], sizeof(i_unsigned) );
            sprintf( buf, "%u", i_unsigned );
            value = buf;
            value_len = strlen( buf );
            break;
         case I_LONG:
            memcpy( &i_long, fields[i], sizeof(i_long) );
            sprintf( buf, "%ld", i_long );
            value = buf;
            value_len = strlen( buf );
            break;
         case I_FLOAT:
            memcpy( &i_float, fields[i], sizeof(i_float) );
            sprintf( buf, "%.5f", i_float );
            value = buf;
            value_len = strlen( buf );
            break;
         case I_DOUBLE:
            memcpy( &i_double, fields[i], sizeof(i_double) );
            sprintf( buf, "%.5f", i_double );
            value = buf;
            value_len = strlen( buf );
            break;
         case I_BINARY:
            value = fields[i];
            value_len = igetfldlen( db, i );
            break;
         default:
            rc = I_ERROR;
            value_len = 0;
            value = NULL;
            break;
      }
      if ( rc == I_ERROR )
         break;
      else
      {
         sprintf( ptr1, "%s%d", ARGPTR(2), i+1 );
         make_upper( ptr1 );
         if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr1, strlen( ptr1 ), value, value_len ) )
         {
            rc = I_ERROR;
            RxSetInternalError( "Error setting Rexx variable: %s", ptr1 );
            break;
         }
      }
   }
   if ( rc != I_ERROR )
   {
      sprintf( ptr1, "%s0", ARGPTR(2) );
      make_upper( ptr1 );
      sprintf( buf1, "%d", num_fields );
      if ( SetRexxVariable( RexxCDTData.RxPackageGlobalData, ptr1, strlen( ptr1 ), buf1, strlen( buf1 ) ) )
      {
         RxSetInternalError( "Error setting Rexx variable: %s", ptr1 );
         rc = I_ERROR;
      }
   }
   free( ptr );
   free( ptr1 );
   free( fields );
   return RxReturnState( retstr, rc ) ;
}

RFH_RETURN_TYPE CDTopen_index
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   Db_Obj *db;
   Index_Obj *idx;
   char *ptr;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 1, 2 ) )
      return( 1 );
   db = arg2db( argv );
   if ( argc == 1
   ||   ARGLEN(1) == 0 )
   {
      ptr = "Physical Index";
   }
   else
   {
      ARGTERM(1);
      ptr = ARGPTR(1);
   }
   idx = ihandle( db, ptr );
   if ( idx == NULL )
      return RxReturnState( retstr, I_ERROR ) ;
   else
      return RxReturnNumber( RexxCDTData.RxPackageGlobalData, retstr, (long)idx ) ;
}

RFH_RETURN_TYPE CDTcreate_index
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   Db_Obj *db;
   int blksize,overwrite,rc;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 3, 3 ) )
      return( 1 );
   db = arg2db( argv );
   blksize = arg2int( argv+1 );
   if ( ARGLEN(2) == 9
   && memcmpi( ARGPTR(2), "overwrite", 9 ) == 0 )
   {
      overwrite = I_YES;
   }
   else if ( ARGLEN(2) == 6
   && memcmpi( ARGPTR(2), "ignore", 6 ) == 0 )
   {
      overwrite = I_NO;
   }
   else
   {
      RxSetInternalError( "Invalid value '%s' for argument 2. Must be one of 'overwrite' or 'ignore'", ARGPTR(2) );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   rc = inew_index_file( db, blksize, overwrite );
   return RxReturnState( retstr, rc ) ;
}

RFH_RETURN_TYPE CDTadd_index
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   Db_Obj *db;
   int i,rc,num_fields;
   char **fielddesc;
   RXSTRING rxs, *prxs;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 3, 3 ) )
      return( 1 );
   db = arg2db( argv );
   ARGTERM(1);
   prxs = GetRexxVariable( RexxCDTData.RxPackageGlobalData,  ARGPTR(2), &rxs, 0 );
   if ( prxs == NULL )
   {
      RxSetInternalError( "Error getting Rexx variable: %s%d", ARGPTR(2), 0 );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   RXSTRTERM( rxs );
   if ( ( num_fields = atoi( RXSTRPTR(rxs) ) ) == 0 )
   {
      free( RXSTRPTR(rxs) );
      return(1);
   }
   free( RXSTRPTR(rxs) );
   /*
    * Allocate space for field pointers.
    */
   if ( ( fielddesc = (char **)malloc( ( num_fields + 1 ) * sizeof(char *) ) ) == NULL )
      return RxReturnState( retstr, I_ERROR ) ;

   for ( i = 0; i < num_fields; i++ )
   {
      prxs = GetRexxVariable( RexxCDTData.RxPackageGlobalData,  ARGPTR(2), &rxs, i+1 );
      if ( prxs == NULL )
      {
         free( fielddesc );
         free( RXSTRPTR(rxs) );
         RxSetInternalError( "Error getting Rexx variable: %s%d", ARGPTR(2), i+1 );
         return RxReturnState( retstr, I_ERROR ) ;
      }
      RXSTRTERM( rxs );
      fielddesc[i] = RXSTRPTR(rxs);
   }
   fielddesc[i] = NULL;
   rc = imkindex( db, ARGPTR(1), fielddesc );
   for ( i = 0; i < num_fields; i++ )
      free( fielddesc[i] );
   free( fielddesc );
   return RxReturnState( retstr, rc ) ;
}

RFH_RETURN_TYPE CDTmodrec
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{ 
   int i,rc,num_fields;
   Db_Obj *db;
   Index_Obj *idx;
   char **recfields;
   RXSTRING rxs, *prxs;
   int *fields;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 3, 3 ) )
      return( 1 );
   db = arg2db( argv );
   if ( ARGLEN(1) == 0 )
      idx = NULL;
   else
      idx = arg2idx( argv+1 ) ;
   ARGTERM(2);
   /*
    * Allocate space for field pointers.
    */
   num_fields = igetfldcount( db );
   fields = (int *)malloc( num_fields*sizeof(int) );
   for ( i = 0; i < num_fields; i++ )
      fields[i] = i;
   prxs = GetRexxVariable( RexxCDTData.RxPackageGlobalData,  ARGPTR(2), &rxs, 0 );
   if ( prxs == NULL )
   {
      free( fields );
      RxSetInternalError( "Error getting Rexx variable: %s%d", ARGPTR(2), 0 );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   RXSTRTERM( rxs );
   if ( atoi( RXSTRPTR(rxs) ) != num_fields )
   {
      free( fields );
      free( RXSTRPTR(rxs) );
      RxSetInternalError( "Number of fields in stem: %s, does not match number of fields in database file: %d", RXSTRPTR(rxs), num_fields );
      return RxReturnState( retstr, I_ERROR ) ;
   }
   free( RXSTRPTR(rxs) );
   rc = I_OK;
   recfields = Stem2CArray( ARGPTR(2), &num_fields, fields, db );
   free( fields );
   if ( recfields == NULL )
      rc = I_ERROR;
   else
   {
      rc = imodrec( db, idx, recfields );
   }
   free_data( num_fields, recfields );
   return RxReturnState( retstr, rc ) ;
}

RFH_RETURN_TYPE CDTopen_db
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{ 
   int shared,readonly;
   Db_Obj *db;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 3, 3 ) )
      return( 1 );
   ARGTERM(0);

   if ( ARGLEN(1) == 6
   && ( memcmpi( ARGPTR(1), "shared", 6 ) == 0 ) )
   {
      shared = I_YES;
   }
   else if ( ARGLEN(1) == 9
   && ( memcmp( ARGPTR(1), "exclusive", 9 ) == 0 ) )
   {
      shared = I_NO;
   }
   else
   {
      RxSetInternalError( "Invalid value '%s' for argument 1. Must be one of 'shared' or 'exclusive'", ARGPTR(1) );
      return RxReturnState( retstr, I_ERROR ) ;
   }

   if ( ARGLEN(2) == 8
   && ( memcmp( ARGPTR(2), "readonly", 8 ) == 0 ) )
   {
      readonly = I_YES;
   }
   else if ( ARGLEN(2) == 9
   && ( memcmp( ARGPTR(2), "readwrite", 9 ) == 0 ) )
   {
      readonly = I_NO;
   }
   else
   {
      RxSetInternalError( "Invalid value '%s' for argument 2. Must be one of 'readonly' or 'readwrite'", ARGPTR(2) );
      return RxReturnState( retstr, I_ERROR ) ;
   }

   db = iopen_db_shared( ARGPTR(0), shared, readonly );
   if ( db == NULL )
      return RxReturnState( retstr, I_ERROR ) ;
   else
      return RxReturnNumber( RexxCDTData.RxPackageGlobalData, retstr, (long)db ) ;
}

RFH_RETURN_TYPE CDTisamexit
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{ 
   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 0, 0 ) )
      return( 1 );
   return RxReturnState( retstr, isamexit() ) ;
}

RFH_RETURN_TYPE CDTisaminit
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{ 
   int bufcount, bufsize;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 2, 2 ) )
      return( 1 );
   bufcount = arg2int( argv+1 );
   bufsize = arg2int( argv+2 );
   return RxReturnState( retstr, isaminit( bufcount, bufsize ) ) ;
}

/*====== Routines for handling registration of functions ======*/


RFH_RETURN_TYPE CDTLoadFuncs
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   int rc = 0L;

#if defined(DYNAMIC_LIBRARY)
   RexxCDTData.RxPackageGlobalData = InitRxPackage( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, &rc );
   if ( !QueryRxFunction( RexxCDTData.RxPackageGlobalData, "CDTISAMINIT" ) )
   {
      /* 
       * Register all external functions
       */
      if ( !rc )
         rc = RegisterRxFunctions( RexxCDTData.RxPackageGlobalData, RexxCDTFunctions, RxPackageName );
   }
#endif
   return RxReturnNumber( RexxCDTData.RxPackageGlobalData, retstr, rc );
}

RFH_RETURN_TYPE CDTDropFuncs
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   RFH_RETURN_TYPE rc=0;
   int unload=0;

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, ( char* )name, argc, argv );
   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 0, 1 ) )
      return( 1 );
   if ( argv[0].strlength == 6
   &&   memcmpi( argv[0].strptr, "UNLOAD", 6 ) == 0 )
      unload = 1;
   rc = ( RFH_RETURN_TYPE )TermRxPackage( &RexxCDTData.RxPackageGlobalData, RexxCDTTerminator, RexxCDTFunctions, RxPackageName, unload );
   return RxReturnNumber( NULL, retstr, rc );
}


RFH_RETURN_TYPE CDTVariable
   (RFH_ARG0_TYPE name, RFH_ARG1_TYPE argc, RFH_ARG2_TYPE argv, RFH_ARG3_TYPE stck, RFH_ARG4_TYPE retstr)
{
   ULONG rc = 0L;
   char buf[50];

   RexxCDTData.RxPackageGlobalData = FunctionPrologue( RexxCDTData.RxPackageGlobalData, RexxCDTInitialiser, (char *)name, argc, argv );

   RxClearInternalError();
   if ( my_checkparam( RexxCDTData.RxPackageGlobalData, name, argc, 1, 2 ) )
      return( 1 );
   if ( argv[0].strlength == 5
   && memcmpi( "debug", argv[0].strptr, 5 ) == 0 )
   {
      if ( argc == 1 )
      {
         return RxReturnNumber( RexxCDTData.RxPackageGlobalData, retstr, RxGetRunFlags( RexxCDTData.RxPackageGlobalData ) );
      }
      else
      {
         RxSetRunFlags ( RexxCDTData.RxPackageGlobalData, atoi( (char *)argv[1].strptr ) );
         return RxReturnNumber( RexxCDTData.RxPackageGlobalData, retstr, 0 );
      }
   }
   else if ( argv[0].strlength == 7
   && memcmpi( "version", argv[0].strptr, 7 ) == 0 )
   {
      if ( argc == 1 )
      {
         sprintf( buf, "%s %s %s", RxPackageName, REXXCDT_VERSION, REXXCDT_DATE );
         return RxReturnString( RexxCDTData.RxPackageGlobalData, retstr, buf );
      }
      else
         return RxReturnString( RexxCDTData.RxPackageGlobalData, retstr, "ERROR: Cannot set variable; VERSION" );
   }
   else if ( argv[0].strlength == 9
   && memcmpi( "debugfile", argv[0].strptr, 9 ) == 0 )
   {
      if ( argc == 1 )
         return RxReturnString( RexxCDTData.RxPackageGlobalData, retstr, RxGetTraceFile( RexxCDTData.RxPackageGlobalData ) );
      else
      {
         rc = RxSetTraceFile( RexxCDTData.RxPackageGlobalData, (char *)argv[1].strptr );
         return RxReturnNumber( RexxCDTData.RxPackageGlobalData, retstr, rc );
      }
   }
   else if ( argv[0].strlength == 8
   && memcmpi( "decimals", argv[0].strptr, 8 ) == 0 )
   {
      if ( argc == 1 )
         return RxReturnNumber( RexxCDTData.RxPackageGlobalData, retstr, RxGetDecimals() );
      else
      {
         RxSetDecimals( atoi( (char *)argv[1].strptr ) );
         return RxReturnNumber( RexxCDTData.RxPackageGlobalData, retstr, 0 );
      }
   }
   sprintf( buf, "ERROR: Invalid variable; %s", argv[0].strptr );
   return RxReturnString( RexxCDTData.RxPackageGlobalData, retstr, buf );
}


/*
 * The following functions are used in rxpackage.c
 */

/*-----------------------------------------------------------------------------
 * Print a usage message.
 *----------------------------------------------------------------------------*/
void RexxCDTUsage

#ifdef HAVE_PROTO
   (void)
#else
   ()
#endif

{
   (void)fprintf(stderr,
      "\nVersion: %s %s %s\n\nUsage:   %s [-h]\n         %s [-idvf<trace file>] [Rexx program name]\n\n",
      RxPackageName,
      REXXCDT_VERSION,
      REXXCDT_DATE,
      RxPackageName,
      RxPackageName);
   exit( 1 );
}

#if defined(USE_REXX6000)
/*
 * This function is used as the entry point for the REXX/6000
 * interpreter
 * If you change this table, don't forget to change the table at the
 * start of this file.
 */
USHORT InitFunc( RXFUNCBLOCK **FuncBlock )
{
   static RXFUNCBLOCK funcarray[] =
   {
      { "CDTADDREC"          , CDTaddrec           ,NULL },
      { "CDTCLOSE_DB"        , CDTclose_db         ,NULL },
      { "CDTCREATE_DB"       , CDTcreate_db        ,NULL },
      { "CDTDELREC"          , CDTdelrec           ,NULL },
      { "CDTFINDHEAD"        , CDTfindhead         ,NULL },
      { "CDTFINDKEY"         , CDTfindkey          ,NULL },
      { "CDTFINDNEXT"        , CDTfindnext         ,NULL },
      { "CDTFINDPREV"        , CDTfindprev         ,NULL },
      { "CDTFINDTAIL"        , CDTfindtail         ,NULL },
      { "CDTGETINDEXDESC"    , CDTgetindexdesc     ,NULL },
      { "CDTGETERROR"        , CDTgeterror         ,NULL },
      { "CDTGETFLDCOUNT"     , CDTgetfldcount      ,NULL },
      { "CDTGETFLDNAMES"     , CDTgetfldnames      ,NULL },
      { "CDTGETFLDTYPE"      , CDTgetfldtype       ,NULL },
      { "CDTGETINDEXNAMES"   , CDTgetindexnames    ,NULL },
      { "CDTGETREC"          , CDTgetrec           ,NULL },
      { "CDTOPEN_INDEX"      , CDTopen_index       ,NULL },
      { "CDTCREATE_INDEX"    , CDTcreate_index     ,NULL },
      { "CDTADD_INDEX"       , CDTadd_index        ,NULL },
      { "CDTMODREC"          , CDTmodrec           ,NULL },
      { "CDTOPEN_DB"         , CDTopen_db          ,NULL },
      { "CDTISAMEXIT"        , CDTisamexit         ,NULL },
      { "CDTISAMINIT"        , CDTisaminit         ,NULL },
      { "CDTLOADFUNCS"       , CDTLoadFuncs        ,NULL },
      { "CDTDROPFUNCS"       , CDTDropFuncs        ,NULL },
      { "CDTVARIABLE"        , CDTVariable         ,NULL },
      { NULL, NULL, NULL }
   } ;
   *FuncBlock = funcarray;
   return (USHORT)0;
}
#endif

void *getRexxCDTFunctionAddress( char *name )
{
   int i, size = sizeof( RexxCDTFunctions ) / sizeof( RexxFunction ) ;

   for (i = 0; i < size; i++)
   {
      if ( strcmp( RexxCDTFunctions[i].InternalName, name) == 0 )
         return RexxCDTFunctions[i].EntryPoint;
   }
   return NULL;
}
